/* topology.cc
 * Copyright (C) 2001 The University of Newcastle upon Tyne
 *
 * Written by Paul Coates (Paul.Coates@ncl.ac.uk)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <arpa/inet.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <netdb.h>
#include "topology.h"
#include "scheduler.h"
#include "discover.h"
#include "preferences.h"
#include "process.h"
#include "task.h"
#include "main.h"
#include "session.h"
#include "data.h"
#include "misc.h"
#include "gui.h"
#include "parser.h"
#include "monitor.h"

Topology::Topology(Main *m)
{
	main = m;
	parser = new Parser();
}

Topology::~Topology()
{
	if (parser!=NULL) delete parser;
}

//////////////////////////////////////////////////////////////////////////////
//
// Code to parse a topology file
//
//////////////////////////////////////////////////////////////////////////////

void Topology::create_placeholder(char *ipstr)
{
	Device *d;
	struct hostent *h;
	struct in_addr addr;
	IPAddress tip;

	tip.setstring(ipstr);

	if (inet_aton(tip.getstring(), &addr) == 0)
	{
		printf("Invalid IP address\n");
		return;
	}

	if ((!tip.zero()) && (!tip.loopback()))
	{
		// see if device already exists
		d = main->session->data->find_ip_on_device(tip);
		if (d==NULL)
		{
			d = new Device();
			d->pingname.set(tip);
			strcpy(d->name, d->pingname.getstring());
			h = gethostbyaddr((const char *) &addr.s_addr, sizeof(addr), AF_INET);
			if (h) strcpy(d->name, h->h_name);
			d->type = 'y';
			main->session->data->devices.tag(d);
			main->gui->update_views();
			main->gui->update_lists();
		}
	}
}

// this is the only in from the rest of the application

void Topology::parse(char *fnameparam, int hop)
{
	char buf[256], fname[256];

	// parse the SNMP data

	sscanf(fnameparam, "%[^ ] %[^\n]", fname, buf);
	sprintf(buf, "Parsing %s\n", fname);
	main->gui->monitor->show_message(buf);
	sprintf(buf, "%s/topology/%s", main->session->datadir, fname);
	parser->load_device(buf, fname);
	if (parser->d==NULL) return;

	// create new device

	tmpdev = new Device();
	origdev = NULL;

	copyDeviceHeader();							// copy over device header from parser
	clearDeviceTags();
	checkForDuplicateDevice();			// identify device that is being replaced

	// if not replacing a device find one to merge with instead
	if (origdev==NULL) joinLayers();

	// There are 8 different combinations for merging devices so make
	// things easier we use a merge_type variable
	mergeType();

	// merge origdev and tmpdev
	attachInterfaces();

	// lets attach at layer 3
	layer3(hop);

	copyOverMachines();

	// now attach at layer 2
	layer2();

	// remove machine entries that are interfaces for the new device
	checkForFalseMachines();

	// we have delayed adding tmpdev to the list of devices until now
	main->session->data->devices.tag(tmpdev);

	deleteTaggedDevices();

	main->gui->update_views();
	main->gui->update_lists();
}

void Topology::copyDeviceHeader()
{
	Link *ln;
	MACAddress *mac, *tmac;

	// copy over stuff from parser->d
	tmpdev->type = parser->d->type;

	tmpdev->pingname = parser->d->pingname;
	tmpdev->pingname2 = parser->d->pingname2;
	strcpy(tmpdev->name, parser->d->name);
	tmpdev->last_seen = parser->d->last_seen;
	strcpy(tmpdev->location, parser->d->location);
	strcpy(tmpdev->community, parser->d->community);
	strcpy(tmpdev->community2, parser->d->community2);
	strcpy(tmpdev->descr, parser->d->descr);
	strcpy(tmpdev->comment, parser->d->comment);
	strcpy(tmpdev->contact, parser->d->contact);
	// default status of unknown
	// default group which we should get from preferences
	// default grab topology
	// default grab statistics
	for (ln=parser->d->loopbacks.get_head(); ln!=NULL; ln=ln->next)
	{
		mac = (MACAddress*) ln->item;

		tmac = new MACAddress();
		tmac->set(*mac);
		tmpdev->loopbacks.tag(tmac);
	}
	tmpdev->merged = 0;
}

// check for machines that match the IP addresses on the interfaces
// we can just delete these machines, we only need to keep them if
// they are MAC address values for layer 2 code

void Topology::clearDeviceTags()
{
	Link *ln;
	Device *d;

	for (ln = main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		d->tag = 0;
	}
}

void Topology::checkForDuplicateDevice()
{
	Link *ln, *ln2;
	Interface *in;
	IPAddress *tip;

	checkForDuplicateDeviceWithIP(&parser->d->pingname);
	if (!parser->d->pingname2.zero())
		checkForDuplicateDeviceWithIP(&parser->d->pingname2);
	for (ln=parser->d->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		for (ln2=in->ip_addresses.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			tip = (IPAddress*) ln2->item;
			checkForDuplicateDeviceWithIP(tip);
		}
	}
}

void Topology::checkForDuplicateDeviceWithIP(IPAddress *ip)
{
	Link *ln;
	Device *d;

	// does the IP match a device
	for (ln = main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		if (d->pingname.compare(*ip)==0)
		{
			// we don't care if placeholder or real device, just copy
			// coordinates and tag for deletion
			if (d->type != 'y') origdev = d;
			tmpdev->x = d->x;
			tmpdev->y = d->y;
			tmpdev->selected = d->selected;
			tmpdev->show = d->show;
			d->tag = 1;
		}
	}
}

void Topology::joinLayers()
{
	Link *ln, *ln2, *ln3;
	Device *d;
	Interface *in;
	MACAddress *mac;

	if (tmpdev->type == 'r')
	{
		for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
		{
			d = (Device*) ln->item;
			if (d->type == 's')
			{
				// printf("Checking switch %s\n", d->name);
				for (ln2=d->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					in = (Interface*) ln2->item;
					// printf("  Interface %d %s\n", in->index, in->mac.getstring());
					for (ln3=tmpdev->loopbacks.get_head(); ln3!=NULL; ln3=ln3->next)
					{
						mac = (MACAddress*) ln3->item;
						// printf("    Router loopback address %s\n", mac->getstring());
						if (mac->compare(in->mac)==1)
						{
							printf("FOUND 2/3 MATCH %s %s\n", tmpdev->name, d->name);
							origdev = d;
							d->tag = 1;
							tmpdev->x = d->x;
							tmpdev->y = d->y;
							tmpdev->selected = d->show;
							tmpdev->show = d->show;
						}
					}
				}
			}
		}
	}else
	{
		for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
		{
			d = (Device*) ln->item;
			if (d->type == 'r')
			{
				for (ln2=parser->d->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					in = (Interface*) ln2->item;
					// printf("  Interface %d %s\n", in->index, in->mac.getstring());
					for (ln3=d->loopbacks.get_head(); ln3!=NULL; ln3=ln3->next)
					{
						mac = (MACAddress*) ln3->item;
						// printf("    Router loopback address %s\n", mac->getstring());
						if (mac->compare(in->mac)==1)
						{
							printf("FOUND 2/3 MATCH %s %s\n", tmpdev->name, d->name);
							origdev = d;
							d->tag = 1;
							tmpdev->x = d->x;
							tmpdev->y = d->y;
							tmpdev->selected = d->show;
							tmpdev->show = d->show;
						}
					}
				}
			}
		}
	}
}

void Topology::mergeType()
{
	if ((tmpdev->type == 'r') && (origdev == NULL))
		merge_type = 1;
	else
	if ((tmpdev->type == 's') && (origdev == NULL))
		merge_type = 2;
	else
	if ((tmpdev->type == 'r') && (origdev != NULL) &&
			(origdev->type == 'r') && (origdev->merged == 0))
		merge_type = 3;
	else
	if ((tmpdev->type == 's') && (origdev != NULL) &&
			(origdev->type == 's') && (origdev->merged == 0))
		merge_type = 4;
	else
	if ((tmpdev->type == 'r') && (origdev != NULL) &&
			(origdev->type == 's') && (origdev->merged == 0))
		merge_type = 5;
	else
	if ((tmpdev->type == 's') && (origdev != NULL) &&
			(origdev->type == 'r') && (origdev->merged == 0))
		merge_type = 6;
	else
	if ((tmpdev->type == 'r') && (origdev != NULL) && (origdev->merged == 1))
		merge_type = 7;
	else
	if ((tmpdev->type == 's') && (origdev != NULL) && (origdev->merged == 1))
		merge_type = 8;
	else
		merge_type = 0;

	if ((merge_type>=5) && (merge_type<=8)) tmpdev->merged = 1;
}

//
// FIXME: may need to map over some layer 3 stuff like speeds to interfaces
//
void Topology::attachInterfaces()
{
	Link *ln, *ln2;
	Interface *in, *tin, *in2;
	Network *n;
	IPAddress *ip, *tip;
	Device *d;
	int done_network;

	d = parser->d;
	if ((merge_type == 5) || (merge_type == 7)) d = origdev;

	if (origdev!=NULL)
	{
		if ((merge_type == 6) || (merge_type == 8))
		{
			strcpy(tmpdev->name, origdev->name);
			tmpdev->type = 'r';
			tmpdev->pingname2 = tmpdev->pingname;
			tmpdev->pingname = origdev->pingname;
			strcpy(tmpdev->community2, tmpdev->community);
			strcpy(tmpdev->community, origdev->community);
		}else
		if ((merge_type == 6) || (merge_type == 8))
		{
			tmpdev->pingname2 = origdev->pingname2;
			strcpy(tmpdev->community2, origdev->community2);
		}
		if (strcmp(origdev->location, "")!=0)
			strcpy(tmpdev->location, origdev->location);
		if (strcmp(origdev->descr, "")!=0)
			strcpy(tmpdev->descr, origdev->descr);
		if (strcmp(origdev->comment, "")!=0)
			strcpy(tmpdev->comment, origdev->comment);
		if (strcmp(origdev->contact, "")!=0)
			strcpy(tmpdev->contact, origdev->contact);
		tmpdev->status = origdev->status;
		tmpdev->group = origdev->group;
		tmpdev->grab_top = origdev->grab_top;
		tmpdev->grab_stats = origdev->grab_stats;
	}

	for (ln=d->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;

		tin = new Interface();
		tin->index = in->index;
		tin->speed = in->speed;
		tin->mac = in->mac;
		tin->type = in->type;
		tin->status = in->status;
		strcpy(tin->descr, in->descr);
		strcpy(tin->comment, in->comment);
		if ((merge_type != 5) && (merge_type != 7))
		{
			for (ln2=in->ip_addresses.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				ip = (IPAddress*) ln2->item;

				tip = new IPAddress();
				tip->set(*ip);
				tin->ip_addresses.tag(tip);
			}
		}
		tin->device = tmpdev;
		tmpdev->interfaces.tag(tin);


		done_network = 0;
		if ((merge_type == 3) || (merge_type == 4))
		{
			// use origdev networks
			in2 = lookupInterface(origdev, tin);
			if (in2 != NULL)
			{
				n = in2->network;
				// since router use new routes to connect up network
				if (merge_type == 3) severNetwork(n);
				n->interfaces.tag(tin);
				tin->network = n;
				done_network = 1;
			}
		}else
		if ((merge_type == 5) || (merge_type == 7))
		{
			in2 = lookupInterface(parser->d, tin);
			if (in2 != NULL)
			{
				tin->speed = in2->speed;
				tin->type = in2->type;
				tin->status = in2->status;
				strcpy(tin->comment, in2->comment);
				for (ln2=in2->ip_addresses.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					ip = (IPAddress*) ln2->item;
					tip = new IPAddress();
					tip->set(*ip);
					tin->ip_addresses.tag(tip);
				}
				n = in->network;
				// since router use new routes to connect up network
				severNetwork(n);
				n->interfaces.tag(tin);
				tin->network = n;
				done_network = 1;
			}
		}else
		if ((merge_type == 6) || (merge_type == 8))
		{
			in2 = lookupInterface(origdev, tin);
			if (in2 != NULL)
			{
				tin->speed = in2->speed;
				tin->type = in2->type;
				tin->status = in2->status;
				strcpy(tin->comment, in2->comment);
				for (ln2=in2->ip_addresses.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					ip = (IPAddress*) ln2->item;
					tip = new IPAddress();
					tip->set(*ip);
					tin->ip_addresses.tag(tip);
				}
				n = in2->network;
				n->interfaces.tag(tin);
				tin->network = n;
				done_network = 1;
			}
		}

		// 1, 2, or those that don't find match
		if (done_network == 0)
		{
			n = new Network();
			n->interfaces.tag(tin);
			tin->network = n;
			main->session->data->networks.tag(n);
		}
	}
}

void Topology::severNetwork(Network *nw)
{
	Link *ln;
	Interface *in;
	Network *n;

	for (ln=nw->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		n = new Network();
		n->interfaces.tag(in);
		in->network = n;
		main->session->data->networks.tag(n);
	}
	nw->interfaces.empty();
}

Interface *Topology::lookupInterface(Device *d, Interface *in)
{
	Interface *in2;
	Link *ln;
	int k, dvlan, ivlan;

	if (in==NULL) return NULL;

	for (ln=d->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in2 = (Interface*) ln->item;
		if ((!in->mac.zero()) && (in2->mac.compare(in->mac)==1))
			return in2;
		if (strcmp(in2->descr, in->descr)==0)
			return in2;
		dvlan = ivlan = -1;
		if (sscanf(in2->descr, "VLAN %d", &k)==1) dvlan = k;
		if (sscanf(in2->descr, "VLAN-%d", &k)==1) dvlan = k;
		if (sscanf(in2->descr, "Vlan%d", &k)==1) dvlan = k;
		if (sscanf(in2->descr, "Vl%d", &k)==1) dvlan = k;
		if (sscanf(in->descr, "VLAN %d", &k)==1) ivlan = k;
		if (sscanf(in->descr, "VLAN-%d", &k)==1) ivlan = k;
		if (sscanf(in->descr, "Vlan%d", &k)==1) ivlan = k;
		if (sscanf(in->descr, "Vl%d", &k)==1) ivlan = k;
		if ((ivlan!=-1) && (ivlan==dvlan))
		{
			return in2;
		}
	}

	return NULL;
}

void Topology::layer3(int hop)
{
	SubnetExtra *r;
	Interface *in, *newin;
	Link *ln, *ln2;
	Subnet *s;

	for (ln=parser->d->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		newin = lookupInterface(tmpdev, in);
		if (newin!=NULL)
		{
			for (ln2=parser->subnets.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				r = (SubnetExtra*) ln2->item;
				if ((r->index == in->index) &&
						(!(r->next.in_range("127.0.0.0", "127.0.0.255"))) &&
						(r->next.zero()==0) && (r->ip.zero()==0))
				{
					if (r->type == 3)
					{
						s = new Subnet();
						s->ip.set(r->ip);
						s->mask = r->mask;
						in->network->subnets.tag(s);
						add_subnet(s);
					}else
					if (r->type == 4)
					{
//		printf("SubnetExtra %s/%d, ", r->ip.getstring(), r->mask);
//		printf("%s, %d, %d\n", r->next.getstring(), r->index, r->type);
						add_layer3(newin->network, r->next, hop);
					}
				}
			}
		}else
			printf("Could not map interface %d %s %s\n", in->index, in->mac.getstring(), in->descr);
	}
}

// adds subnet to global list if not already added
void Topology::add_subnet(Subnet *s)
{
	Link *ln;
	Subnet *s2;
	int got=0;

	if (s->mask > 30) return;
	for (ln=main->session->data->subnets.get_head(); ln!=NULL; ln=ln->next)
	{
		s2 = (Subnet*) ln->item;
		if ((s->ip.compare(s2->ip)==0) && (s2->mask == s->mask)) got++;
	}
	if (!got) main->session->data->subnets.tag(s);
}

void Topology::add_layer3(Network *n, IPAddress ip, int hop)
{
	Link *ln, *ln2, *ln3;
	char buf[1024];
	Device *d;
	Interface *in;
	IPAddress *tip;
	int found;

	if (tmpdev->type != 'r') return;

	found = 0;
	for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		for (ln2=d->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			in = (Interface*) ln2->item;
			for (ln3=in->ip_addresses.get_head(); ln3!=NULL; ln3=ln3->next)
			{
				tip = (IPAddress*) ln3->item;
				if (tip->compare(ip)==0)
				{
					if (in->network != n)
					{
						// we need to merge the networks n & in->network
						merge_networks(n, in->network);
					}
					found = 1;
				}
			}
		}
	}

	// layer 3 discovery
	if ((found==0) && (hop>0))
	{
		// need to make sure that task is not already assigned &&
		// the ip passes the discovery filter

		if (main->gui->discover->match(ip, hop-1))
		{
			sprintf(buf, "%s %s", ip.getstring(), get_community_list_string());
			main->scheduler->addTask(buf, 1, 3, hop-1);
		}
	}
}

void Topology::merge_networks(Network *n1, Network *n2)
{
	Link *ln;
	Interface *in;
	Machine *m;

	if (n1==n2) return;

	// n1 is the new network
	// n2 is the old network
	// we move everything from n2 to n1

	// move over interfaces
	for (ln=n2->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		if (in!=NULL)
		{
			in->network = n1;
			n1->interfaces.tag(in);
		}
	}

	// now move over subnets (avoid duplicates)

	// now move over machines (avoid duplicates)
	for (ln=n2->machines.get_head(); ln!=NULL; ln=ln->next)
	{
		m = (Machine*) ln->item;
		if (m!=NULL)
		{
			m->network = n1;
			n1->machines.tag(m);
		}
	}

	main->session->data->networks.del(n2);
	if (main->gui->monitor->viewer->cur == n2)
		main->gui->monitor->viewer->cur = NULL;
	delete n2;			// DND
}

void Topology::layer2()
{
	Link *ln, *ln2;
	Device *d;
	Interface *in;

	if (tmpdev->type != 's') return;
	printf("LAYER 2, we have a switch. On this switch we have the following\n");

	for (ln=main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		if (d->type == 'r')
		{
printf("  Checking device %s\n", d->name);
			for (ln2=d->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				in = (Interface*) ln2->item;
printf("    Checking interface %s %s\n", in->descr, in->mac.getstring());
				if (!in->mac.zero())
				{
					layer2_check(in);
					printf("Check for mac address %s %s %s in tmpdev\n", in->device->name, in->descr, in->mac.getstring());
				}
			}
		}
	}
}

void Topology::layer2_check(Interface *tin)
{
	Link *ln, *ln2;
	Interface *in;
	Machine *m;

	for (ln=tmpdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
//		printf("INTERFACE %s\n", in->descr);
		for (ln2 = in->network->machines.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			m = (Machine*) ln2->item;
//printf("  Machine %s\n", m->mac.getstring());
			if (m->mac.compare(tin->mac)==1)
			{
				printf("FOUND a place to attach switch %s - %s/%s\n",
					tmpdev->name, tin->device->name, tin->descr);
				merge_networks(in->network, tin->network);
			}
		}
	}
}

void Topology::checkForFalseMachines()
{
	Link *ln, *ln2;
	Interface *in;
	IPAddress *tip;

	checkForFalseMachinesWithIP(&parser->d->pingname);
	if (!parser->d->pingname2.zero())
		checkForFalseMachinesWithIP(&parser->d->pingname2);
	for (ln=parser->d->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		in = (Interface*) ln->item;
		for (ln2=in->ip_addresses.get_head(); ln2!=NULL; ln2=ln2->next)
		{
			tip = (IPAddress*) ln2->item;
			checkForFalseMachinesWithIP(tip);
		}
	}
}

void Topology::checkForFalseMachinesWithIP(IPAddress *ip)
{
	Link *ln;
	Machine *m;

	// does the IP match a machine
	for (ln = main->session->data->machines.get_head(); ln!=NULL; ln=ln->next)
	{
		m = (Machine*) ln->item;
		if (m->ip.compare(*ip)==0)
		{
			if (m->network) m->network->machines.del(m);
			ln->item=NULL;
			if (main->gui->monitor->viewer->cur == m)
				main->gui->monitor->viewer->cur = NULL;
			delete m;		// DND
		}
	}
	main->session->data->machines.tidy();
}

void Topology::copyOverMachines()
{
	Link *ln, *ln2;
	Interface *oin, *nin;
	Machine *m;

	for (ln=tmpdev->interfaces.get_head(); ln!=NULL; ln=ln->next)
	{
		nin = (Interface*) ln->item;
		if ((nin->type != 24) &&		// loopback
				(nin->type != 53))			// virtual
		{
			oin = lookupInterface(parser->d, nin);
			if (oin!=NULL)
			{
				for (ln2=oin->network->machines.get_head(); ln2!=NULL; ln2=ln2->next)
				{
					m = (Machine*) ln2->item;
					// do lots of checks here, look for duplicates
					nin->network->machines.tag(m);
					m->network = nin->network;
					main->session->data->machines.tag(m);
				}
			}
		}
	}
}

// delete tagged device
// delete leaf networks
// update drag & drop
void Topology::deleteTaggedDevices()
{
	Link *ln, *ln2;
	Interface *in;
	Device *d;

//	main->gui->monitor->viewer->cur

	for (ln = main->session->data->devices.get_head(); ln!=NULL; ln=ln->next)
	{
		d = (Device*) ln->item;
		if (d->tag==1)
		{
			for (ln2=d->interfaces.get_head(); ln2!=NULL; ln2=ln2->next)
			{
				in = (Interface*) ln2->item;
				if ((in != NULL) && (in->network != NULL))
				{
					in->network->interfaces.del(in);
					if (in->network->interfaces.size()==0)
					{
						main->session->data->networks.del(in->network);
						if (main->gui->monitor->viewer->cur == in->network)
							main->gui->monitor->viewer->cur = NULL;
						delete (in->network);			// DND
					}
				}
			}
			if (main->gui->monitor->viewer->cur == d)
				main->gui->monitor->viewer->cur = NULL;
			delete d;			// DND
			ln->item = NULL;
		}
	}
	main->session->data->devices.tidy();
}

